from sklearn.metrics import log_loss, mean_absolute_error, mean_squared_error, r2_score, f1_score
from abc import ABCMeta, abstractmethod

from ..util.static_types import StaticTypes


class Scorer(object):
    """
    Base Class for all skater scoring functions.
    Any Scoring function must consume a model.
    Any scorer must determine the types of models that are compatible.

    """

    __metaclass__ = ABCMeta


    model_types = None
    prediction_types = None
    label_types = None

    def __init__(self, model):
        self.model = model

    @classmethod
    def check_params(cls):
        assert all([i in StaticTypes.model_types._valid_ for i in cls.model_types])
        assert all([i in StaticTypes.output_types._valid_ for i in cls.prediction_types])
        assert all([i in StaticTypes.output_types._valid_ for i in cls.label_types])


    @classmethod
    def check_model(cls, model):

        assert model.model_type in cls.model_types, "Scorer {0} not valid for models of type {1}, " \
                                                    "only {2}".format(cls,
                                                                      model.model_type,
                                                                      cls.model_types)


    def __call__(self, y_true, y_predicted, sample_weight=None):
        self.check_model(self.model)
        self.check_data(y_true, y_predicted)
        # formatted_y = self.model.transformer(self.model.output_formatter(y_true))
        return self._score(y_true, y_predicted, sample_weight=sample_weight)


    @staticmethod
    @abstractmethod
    def check_data(y_true, y_predicted):
        pass


class RegressionScorer(Scorer):
    model_types = [StaticTypes.model_types.regressor]
    prediction_types = [
        StaticTypes.output_types.numeric,
        StaticTypes.output_types.float,
        StaticTypes.output_types.int
    ]
    label_types = [
        StaticTypes.output_types.numeric,
        StaticTypes.output_types.float,
        StaticTypes.output_types.int
    ]

    @staticmethod
    def check_data(y_true, y_predicted):
        assert hasattr(y_predicted, 'shape'), \
            'outputs must have a shape attribute'
        assert hasattr(y_true, 'shape'), \
            'y_true must have a shape attribute'
        assert (len(y_predicted.shape) == 1) or (y_predicted.shape[1] == 1), \
            "Regression outputs must be 1D, " \
            "got {}".format(y_predicted.shape)
        assert (len(y_true.shape) == 1) or (y_true.shape[1] == 1), \
            "Regression outputs must be 1D, " \
            "got {}".format(y_true.shape)


# Regression Scorers
class MeanSquaredError(RegressionScorer):
    __name__ = "MSE"
    type = StaticTypes.scorer_types.decreasing

    @staticmethod
    def _score(y_true, y_predicted, sample_weight=None):
        return mean_squared_error(y_true, y_predicted, sample_weight=sample_weight)


class MeanAbsoluteError(RegressionScorer):
    __name__ = "MAE"
    type = StaticTypes.scorer_types.decreasing

    @staticmethod
    def _score(y_true, y_predicted, sample_weight=None):
        return mean_absolute_error(y_true, y_predicted, sample_weight=sample_weight)


class RSquared(RegressionScorer):
    __name__ = "R2"
    # Reference: https://en.wikipedia.org/wiki/Coefficient_of_determination
    # The score values range between [0, 1]. The best possible value is 1, however one could expect negative values as
    # well because of the arbitrary model fit.
    type = StaticTypes.scorer_types.increasing

    @staticmethod
    def _score(y_true, y_predicted, sample_weight=None):
        return r2_score(y_true, y_predicted, sample_weight=sample_weight)


class ClassifierScorer(Scorer):

    """
    * predictions must be N x K matrix with N rows and K classes.
    * labels must be be N x K matrix with N rows and K classes.
    """

    model_types = [StaticTypes.model_types.classifier]
    prediction_types = [StaticTypes.output_types.numeric, StaticTypes.output_types.float, StaticTypes.output_types.int]
    label_types = [StaticTypes.output_types.numeric, StaticTypes.output_types.float, StaticTypes.output_types.int]

    @staticmethod
    def check_data(y_true, y_predicted):
        assert hasattr(y_predicted, 'shape'), 'outputs must have a shape attribute'
        assert hasattr(y_true, 'shape'), 'y_true must have a shape attribute'


# Metrics related to Classification
class CrossEntropy(ClassifierScorer):
    __name__ = "cross-entropy"
    type = StaticTypes.scorer_types.decreasing

    @staticmethod
    def _score(y_true, y_predicted, sample_weight=None):
        """

        :param X: Dense X of probabilities, or binary indicator
        :param y:
        :param sample_weights:
        :return:
        """
        return log_loss(y_true, y_predicted, sample_weight=sample_weight)


class F1(ClassifierScorer):
    __name__ = "f1-score"
    type = StaticTypes.scorer_types.increasing

    @staticmethod
    def _score(y_true, y_predicted, sample_weight=None, average='weighted'):
        """

        :param X: Dense X of probabilities, or binary indicator
        :param y: indicator
        :param sample_weights:
        :return:
        """
        if len(y_predicted.shape) == 2:
            preds = y_predicted.argmax(axis=1)
        else:
            preds = y_predicted

        return f1_score(y_true, preds, sample_weight=sample_weight, average=average)


class ScorerFactory(object):
    """
    The idea is that we initialize the object with the model,
    but also provide an api for retrieving a static scoring function
    after checking that things are ok.
    """
    def __init__(self, model):
        if model.model_type == StaticTypes.model_types.regressor:
            self.mse = MeanSquaredError(model)
            self.mae = MeanAbsoluteError(model)
            self.r2 = RSquared(model)
            self.default = self.mae
        elif model.model_type == StaticTypes.model_types.classifier:
            self.cross_entropy = CrossEntropy(model)
            self.f1 = F1(model)
            # TODO: Not sure why the first condition is relevant, probably some early design decision.
            # TODO Need to check other examples and add more test before removing the first check completely
            if model.probability is not None and not 'unknown' or model.probability is True:
                self.default = self.cross_entropy
            else:
                self.default = self.f1

        self.type = self.default.type


    def __call__(self, y_true, y_predicted, sample_weight=None):
        return self.default(y_true, y_predicted, sample_weight=sample_weight)


    def get_scorer_function(self, scorer_type='default'):
        """
        Returns a scoring function as a pure function.

        Parameters
        ----------

        scorer_type: string
            Specifies which scorer to use. Default value 'default' returns f1 for classifiers that return labels,
            cross_entropy for classifiers that return probabilities, and mean absolute error for regressors.


        Returns
        -------
            .score staticmethod of skater.model.scorer.Scorer object.
        """
        assert scorer_type in self.__dict__, "Scorer type {} not recognized " \
                                             "or allowed for model type".format(scorer_type)
        scorer = self.__dict__[scorer_type]._score
        scorer.type = self.__dict__[scorer_type].type
        scorer.name = self.__dict__[scorer_type].__name__
        return scorer



from __future__ import division
import numpy as np
from sklearn.preprocessing import LabelBinarizer
from skater.util.logger import build_logger
from skater.util.logger import _INFO

logger = build_logger(_INFO, __name__)


def flatten(array):
    return [item for sublist in array for item in sublist]


def add_column_numpy_array(array, new_col):
    placeholder = np.ones(array.shape[0])[:, np.newaxis]
    result = np.hstack((array, placeholder))

    if isinstance(new_col, np.ndarray):
        assert array.shape[0] == new_col.shape[0], "input array row counts \
                                                    must be the same. \
                                                    Expected: {0}\
                                                    Actual: {1}".format(array.shape[0],
                                                                        new_col.shape[0])
        assert len(new_col.shape) <= 2, "new column must be 1D or 2D"

        if len(new_col.shape) == 1:
            new_col = new_col[:, np.newaxis]
        return np.hstack((array, new_col))
    elif isinstance(new_col, list):
        assert len(new_col) == array.shape[0], "input array row counts \
                                                    must be the same. \
                                                    Expected: {0}\
                                                    Actual: {1}".format(len(array),
                                                                        len(new_col))
        new_col = np.array(new_col)
        assert len(new_col.shape) == 1, "list elements cannot be iterable"
        new_col = new_col[:, np.newaxis]
        return np.hstack((array, new_col))
    else:
        placeholder = np.ones(array.shape[0])[:, np.newaxis]
        result = np.hstack((array, placeholder))
        result[:, -1] = new_col
        return result


def allocate_samples_to_bins(n_samples, ideal_bin_count=100):
    """goal is as best as possible pick a number of bins
    and per bin samples to a achieve a given number
    of samples.

    Parameters
    ----------

    Returns
    ----------
    number of bins, list of samples per bin
    """

    if n_samples <= ideal_bin_count:
        n_bins = n_samples
        samples_per_bin = [1 for _ in range(n_bins)]
    else:
        n_bins = ideal_bin_count
        remainer = n_samples % ideal_bin_count

        samples_per_bin = np.array([(n_samples - remainer) / ideal_bin_count for _ in range(n_bins)])
        if remainer != 0:
            additional_samples_per_bin = distribute_samples(remainer, n_bins)
            samples_per_bin = samples_per_bin + additional_samples_per_bin
    return n_bins, np.array(samples_per_bin).astype(int)


def distribute_samples(n_samples, n_bins):
    assert n_samples < n_bins, "number of samples should be \
                                less than number of bins"
    space_size = n_bins / n_samples

    samples_per_bin = np.zeros(n_bins).tolist()

    index_counter = 0
    for sample in range(n_samples):
        index = int(index_counter)
        samples_per_bin[index] += 1
        index_counter += space_size
    return np.array(samples_per_bin).astype(int)


def divide_zerosafe(a, b):
    """ diving by zero returns 0 """
    with np.errstate(divide='ignore', invalid='ignore'):
        c = np.true_divide(a, b)
        c[~np.isfinite(c)] = 0  # -inf inf NaN
    return c


# Lambda for converting data-frame to a dictionary
convert_dataframe_to_dict = lambda key_column_name, value_column_name, df: \
    df.set_index(key_column_name).to_dict()[value_column_name]


def json_validator(json_object):
    """ json validator
    """
    # Reference: https://stackoverflow.com/questions/5508509/how-do-i-check-if-a-string-is-valid-json-in-python
    import json
    try:
        json.loads(json_object)
    except ValueError:
        return False
    return True


def _render_html(file_name, width=None, height=None):
    width, height
    from IPython.core.display import HTML
    return HTML(file_name)


def _render_image(file_name, width=600, height=300):
    from IPython.display import Image
    return Image(file_name, width=width, height=height)


def _render_pdf(file_name, width=600, height=300):
    from IPython.display import IFrame
    IFrame(file_name, width=width, height=height)


def show_in_notebook(file_name_with_type='rendered.html', width=600, height=300, mode=None):
    """ Display generated artifacts(e.g. .png, .html, .jpeg/.jpg) in interactive Jupyter style Notebook

    Parameters
    -----------
    file_name_with_type: str
        specify the name of the file to display
    width: int
        width in pixels to constrain the image
    height: int
        height in pixels to constrain the image
    """
    from IPython.core.display import display, HTML
    if mode is not 'interactive':
        file_type = file_name_with_type.split('/')[-1].split('.')[-1]
        choice_dict = {
            'html': _render_html,
            'png': _render_image,
            'jpeg': _render_image,
            'jpg': _render_image,
            'svg': _render_image,
            'pdf': _render_pdf
        }
        select_type = lambda choice_type: choice_dict[file_type]
        logger.info("File Name: {}".format(file_name_with_type))
        return display(select_type(file_type)(file_name_with_type, width, height))
    else:
        # For now using iframe for some interactive plotting. This should be replaced with a better plotting interface
        iframe_style = '<div style="-webkit-overflow-scrolling:touch; overflow-x:hidden; ' \
                       'overflow-y:auto; width:{}px; height:{}px; margin: -1.2em; ' \
                       '-webkit-transform: scale(0.9) -moz-transform-scale(0.5)"> ' \
                       '<iframe src={} style="width:100%; height:100%; frameborder:1px;">' \
                       '</iframe>' \
                       '</div>'.format(width, height, file_name_with_type)
        return HTML(iframe_style)


class MultiColumnLabelBinarizer(LabelBinarizer):
    def __init__(self, neg_label=0, pos_label=1, sparse_output=False):
        self.neg_label = neg_label
        self.pos_label = pos_label
        self.sparse_output = sparse_output
        self.binarizers = []


    def fit(self, X):
        for x in X.T:
            binarizer = LabelBinarizer()
            binarizer.fit(x)
            self.binarizers.append(binarizer)


    def transform(self, X):
        results = []
        for i, x in enumerate(X.T):
            results.append(self.binarizers[i].transform(x))
        return np.concatenate(results, axis=1)


    def fit_transform(self, X):
        self.fit(X)
        return self.transform(X)


    def inverse_transform(self, X):
        results = []
        column_counter = 0

        for i, binarizer in enumerate(self.binarizers):
            n_cols = binarizer.classes_.shape[0]
            x_subset = X[:, column_counter:column_counter + n_cols]
            inv = binarizer.inverse_transform(x_subset)
            if len(inv.shape) == 1:
                inv = inv[:, np.newaxis]
            results.append(inv)
            column_counter += n_cols
        return np.concatenate(results, axis=1)



from sklearn.externals.six import StringIO
from sklearn.tree import export_graphviz
import pydotplus
import numpy as np

from skater.util import exceptions
try:
    import matplotlib
    matplotlib.use('agg')
    import matplotlib.pyplot as plt
    from matplotlib.colors import rgb2hex
except ImportError:
    raise exceptions.MatplotlibUnavailableError("matplotlib is required but unavailable on the system.")


# reference: http://wingraphviz.sourceforge.net/wingraphviz/language/colorname.htm
# TODO: Make the color scheme for regression and classification homogeneous
color_schemes = ['aliceblue', 'antiquewhite', 'aquamarine', 'azure', 'beige', 'bisque', 'black', 'blanchedalmond', 'blue',
                 'blueviolet', 'brown', 'burlywood', 'cadetblue', 'chartreuse', 'chocolate', 'coral', 'cornflowerblue',
                 'cornsilk', 'crimson', 'cyan', 'darkgoldenrod', 'darkgreen', 'darkkhaki', 'darkolivegreen', 'darkorange',
                 'darkorchid', 'darksalmon', 'darkseagreen', 'darkslateblue', 'darkslategray', 'darkslategrey',
                 'darkturquoise', 'darkviolet', 'deeppink', 'deepskyblue', 'dimgray', 'dimgrey', 'dodgerblue', 'firebrick',
                 'floralwhite', 'forestgreen', 'gainsboro', 'ghostwhite', 'gold', 'goldenrod', 'gray', 'green',
                 'greenyellow', 'grey', 'honeydew', 'hotpink', 'indianred', 'indigo', 'ivory', 'khaki', 'lavender',
                 'lawngreen', 'lemonchiffon', 'lightblue', 'lightcoral', 'lightcyan', 'lightgoldenrod',
                 'lightgoldenrodyellow', 'lightgray', 'lightgrey', 'lightpink', 'lightsalmon', 'lightseagreen',
                 'lightskyblue', 'lightslateblue', 'lightslategray', 'lightslategrey', 'lightsteelblue', 'lightyellow',
                 'limegreen', 'linen', 'magenta', 'maroon', 'mediumaquamarine', 'mediumblue', 'mediumorchid',
                 'mediumpurple', 'mediumseagreen', 'mediumslateblue', 'mediumspringgreen', 'mediumturquoise',
                 'mediumvioletred', 'midnightblue', 'mintcream', 'mistyrose', 'moccasin', 'navajowhite', 'navy',
                 'navyblue', 'oldlace', 'orange', 'orangered', 'orchid', 'palegoldenrod', 'palegreen', 'paleturquoise',
                 'palevioletred', 'papayawhip', 'peru', 'pink', 'plum', 'powderblue', 'purple', 'red', 'rosybrown',
                 'royalblue', 'saddlebrown', 'salmon', 'sandybrown', 'seagreen', 'seashell', 'sienna', 'skyblue',
                 'slateblue', 'slategray', 'slategrey', 'snow', 'springgreen', 'steelblue', 'tan', 'thistle', 'tomato',
                 'turquoise', 'violet', 'violetred', 'wheat', 'white', 'whitesmoke', 'yellow', 'yellowgreen']


# Reference: https://github.com/scikit-learn/scikit-learn/blob/a24c8b464d094d2c468a16ea9f8bf8d42d949f84/sklearn/tree/_tree.pyx
TREE_LEAF = -1
TREE_UNDEFINED = -2


def _get_colors(num_classes, random_state=1):
    np.random.seed(random_state)
    color_index = np.random.randint(0, len(color_schemes), num_classes)
    colors = np.array(color_schemes)[color_index]
    return colors


def _generate_graph(est, est_type='classifier', classes=None, features=None,
                    enable_node_id=True, coverage=True):
    dot_data = StringIO()
    # class names are needed only for "Classification" for "Regression" it is set to None
    c_n = classes if est_type == 'classifier' else None
    export_graphviz(est, out_file=dot_data, filled=True, rounded=True,
                    special_characters=True, feature_names=features,
                    class_names=c_n, node_ids=enable_node_id, proportion=coverage)
    graph = pydotplus.graph_from_dot_data(dot_data.getvalue())
    return graph


def _set_node_properites(estimator, estimator_type, graph_instance, color_names, default_color):
    # Query and assign properties to each node
    thresholds = estimator.tree_.threshold
    values = estimator.tree_.value
    left_node = estimator.tree_.children_left
    right_node = estimator.tree_.children_right

    nodes = graph_instance.get_node_list()
    for node in nodes:
        if node.get_name() not in ('node', 'edge'):
            if estimator_type == 'classifier':
                value = values[int(node.get_name())][0]
                # 1. Color only the leaf nodes, where one class is dominant or if it is a leaf node
                # 2. For mixed population or otherwise set the default color
                if max(value) == sum(value) or thresholds[int(node.get_name())] == TREE_UNDEFINED or \
                        left_node[int(node.get_name())] and right_node[int(node.get_name())] == TREE_LEAF:
                    node.set_fillcolor(color_names[np.argmax(value)])
                else:
                    node.set_fillcolor(default_color)
            else:
                # if the estimator type is a "regressor", then the intensity of the color is defined by the
                # population coverage for a particular value
                percent = estimator.tree_.n_node_samples[int(node.get_name())] / float(estimator.tree_.n_node_samples[0])
                rgba = plt.cm.get_cmap(color_names)(percent)
                hex_code = rgb2hex(rgba)
                node.set_fillcolor(hex_code)
                graph_instance.set_colorscheme(color_names)
    return graph_instance


# https://stackoverflow.com/questions/48085315/interpreting-graphviz-output-for-decision-tree-regression
# https://stackoverflow.com/questions/42891148/changing-colors-for-decision-tree-plot-created-using-export-graphviz
# Color scheme info: http://wingraphviz.sourceforge.net/wingraphviz/language/colorname.htm
# Currently, supported only for sklearn models
def plot_tree(estimator, estimator_type='classifier', feature_names=None, class_names=None, color_list=None,
              colormap_reg='PuBuGn', enable_node_id=True, coverage=True, seed=2):

    graph = _generate_graph(estimator, estimator_type, class_names, feature_names, enable_node_id, coverage)

    if estimator_type == 'classifier':
        # if color is not assigned, pick color uniformly random from the color list defined above if the estimator
        # type is "classification"
        colors = color_list if color_list is not None else _get_colors(len(class_names), seed)
        default_color = 'cornsilk'
    else:
        colors = colormap_reg
        default_color = None

    graph = _set_node_properites(estimator, estimator_type, graph, color_names=colors, default_color=default_color)

    # Set the color scheme for the edges
    edges = graph.get_edge_list()
    for ed in edges:
        ed.set_color('steelblue')
    return graph


_return_value = lambda estimator_type, v: 'Predicted Label: {}'.format(str(np.argmax(v))) \
    if estimator_type == 'classifier' else 'Value: {}'.format(str(v))


def _global_decisions_as_txt(est_type, label_color, criteria_color, if_else_color, values,
                             features, thresholds, l_nodes, r_nodes):
    # define "if and else" string patterns for extracting the decision rules
    if_str_pattern = lambda offset, node: offset + "if {}{}".format(criteria_color, features[node]) \
        + " <= {}".format(str(thresholds[node])) + if_else_color + " {"

    other_str_pattern = lambda offset, str_type: offset + if_else_color + str_type

    def _recurse_tree(left_node, right_node, threshold, node, depth=0):
        offset = "  " * depth
        if threshold[node] != TREE_UNDEFINED:
            print(if_str_pattern(offset, node))
            if left_node[node] != TREE_LEAF:
                _recurse_tree(left_node, right_node, threshold, left_node[node], depth + 1)
                print(other_str_pattern(offset, "} else {"))
                if right_node[node] != TREE_LEAF:
                    _recurse_tree(left_node, right_node, threshold, right_node[node], depth + 1)
                print(other_str_pattern(offset, "}"))
        else:
            print(offset, label_color, _return_value(est_type, values[node]))

    _recurse_tree(l_nodes, r_nodes, thresholds, 0)


def _local_decisions_as_txt(est, est_type, label_color, criteria_color, if_else_color,
                            values, features, thresholds, input_X):
    greater_or_less = lambda f_v, s_c: "<=" if f_v <= s_c else ">"
    as_str_pattern = lambda offset, node_id, \
        feature_value, sign: offset + \
        "As {}{}{}".format(criteria_color, features[node_id], "[" + str(feature_value) + "]") + \
        " {} {}".format(sign, str(thresholds[node_id])) + if_else_color + " then,"

    path = est.decision_path(input_X.values.reshape(1, -1))
    node_indexes = path.indices
    leaf_id = est.apply(input_X.values.reshape(1, -1))
    depth = 0
    for node_index in node_indexes:
        offset = "  " * depth
        if leaf_id != node_index:
            feature_value = input_X[features[node_index]]
            print(as_str_pattern(offset, node_index, feature_value,
                                 greater_or_less(feature_value, thresholds[node_index])))
            depth += 1
        else:
            print(offset, label_color, _return_value(est_type, values[node_index]))


# Current implementation is specific to sklearn models.
# Reference: https://stackoverflow.com/questions/20224526/how-to-extract-the-decision-rules-from-scikit-learn-decision-tree
# TODO: Figure out ways to make it generic for other frameworks
def tree_to_text(tree, feature_names, estimator_type='classifier', scope='global', X=None):
    # defining colors
    label_value_color = "\033[1;34;49m"  # blue
    split_criteria_color = "\033[0;32;49m"  # green
    if_else_quotes_color = "\033[0;30;49m"  # if and else quotes

    left_nodes = tree.tree_.children_left
    right_nodes = tree.tree_.children_right
    criterias = tree.tree_.threshold
    feature_names = [feature_names[i] for i in tree.tree_.feature]
    values = tree.tree_.value

    if scope == "global":
        return _global_decisions_as_txt(estimator_type, label_value_color, split_criteria_color,
                                        if_else_quotes_color, values, feature_names, criterias, left_nodes, right_nodes)
    else:
        return _local_decisions_as_txt(tree, estimator_type, label_value_color, split_criteria_color,
                                       if_else_quotes_color, values, feature_names, criterias, X)



"""Funcs for logging"""
import logging


_CRITICAL = logging.CRITICAL
_ERROR = logging.ERROR
_WARNING = logging.WARNING
_INFO = logging.INFO
_DEBUG = logging.DEBUG
_NOTSET = logging.NOTSET


def build_logger(log_level, logger_name, capture_warning=True):
    logger = logging.Logger(logger_name)

    # All warnings are logged by default
    logging.captureWarnings(capture_warning)

    logger.setLevel(log_level)

    msg_formatter = logging.Formatter(
        '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
    )

    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(msg_formatter)
    stream_handler.setFormatter(msg_formatter)
    logger.addHandler(stream_handler)
    return logger



"""Model class."""

import abc
import numpy as np
from sklearn.preprocessing import LabelEncoder, label_binarize
from sklearn.utils.multiclass import type_of_target
import pandas as pd

from ..util.static_types import StaticTypes
from ..util.logger import build_logger
from ..util import exceptions
from ..data import DataManager
from .scorer import RSquared, CrossEntropy, MeanSquaredError, MeanAbsoluteError, ScorerFactory


class ModelType(object):
    """What is a model? A model needs to make predictions, so a means of
    passing data into the model, and receiving results.

    Goals:
        We want to abstract away how we access the model.
        We want to make inferences about the format of the output.
        We want to able to map model outputs to some smaller, universal set of output types.
        We want to infer whether the model is real valued, or classification (n classes?)

    # Todos:
    * check unique_vals are unique
    * check that if probability=False, predictions arent continuous

    """
    __metaclass__ = abc.ABCMeta

    def __init__(self,
                 log_level=30,
                 target_names=None,
                 examples=None,
                 feature_names=None,
                 unique_values=None,
                 input_formatter=None,
                 output_formatter=None,
                 model_type=None,
                 probability=None):
        """
        Base model class for wrapping prediction functions. Common methods
        involve output type inference in requiring predict methods

        Parameters
        ----------
            log_level: int
                0, 10, 20, 30, 40, or 50 for verbosity of logs.
            target_names: arraytype
                The names of the target variable/classes. There should be as many
                 names as there are outputs per prediction (n=1 for regression,
                 n=2 for binary classification, etc). Defaults to Predicted Value for
                 regression and Class 1...n for classification.



        Attributes
        ----------
            model_type: string

        """

        self._check_model_type(model_type)
        self._check_probability(probability)

        self._log_level = log_level
        self.logger = build_logger(log_level, __name__)
        self.examples = None

        self.output_var_type = StaticTypes.unknown
        self.output_shape = StaticTypes.unknown
        self.n_classes = StaticTypes.unknown
        self.input_shape = StaticTypes.unknown
        if probability is None:
            self.probability = StaticTypes.unknown
        else:
            self.probability = probability

        if model_type is None:
            self.model_type = StaticTypes.unknown
        else:
            self.model_type = model_type

        self.transformer = identity_function
        self.label_encoder = LabelEncoder()
        self.target_names = target_names
        self.feature_names = feature_names
        self.unique_values = unique_values
        self.input_formatter = input_formatter or identity_function
        self.output_formatter = output_formatter or identity_function

        self.has_metadata = False

        if examples is not None:
            self.input_type = type(examples)
            examples = DataManager(examples, feature_names=feature_names)
            self._build_model_metadata(examples)
        else:
            self.input_type = None
            self.logger.warn("No examples provided, cannot infer model type")


    def _check_model_type(self, model_type):
        __types__ = [None,
                     StaticTypes.model_types.classifier,
                     StaticTypes.model_types.regressor]
        assert model_type in __types__, \
            "Expected model_type {0}, got {1}".format(__types__, model_type)


    def _check_probability(self, probability):
        __types__ = [None, True, False]
        assert probability in __types__, \
            "Expected model_type {0}, got {1}".format(__types__, probability)


    def predict(self, *args, **kwargs):
        """
        The way in which the submodule predicts values given an input
        """
        if self.has_metadata is False:
            self.has_metadata = True
            examples = DataManager(*args)
            self._build_model_metadata(examples)
        return self.transformer(self.output_formatter(self._execute(self.input_formatter(*args, **kwargs))))


    @property
    def scorers(self):
        """

        :param X:
        :param y:
        :param sample_weights:
        :param scorer:
        :return:
        """
        # TODO: This is a temporary work around. I feel that Scorer component needs to be designed slightly different.
        # Ideally, with the initialization of the Interpretation instance we want to get Interpretation Algorithms all
        # ready to go. Similary, when we initialize InMemory model/Deploy model,
        # we should be able to specifying the evaluation type but giving user the freedom to specify the metric
        # later as well. The below, condition is commented to implement TreeSurrogate.
        # if not self.has_metadata:
        #     raise NotImplementedError("The model needs metadata before "
        #                               "the scorer can be used. Please first"
        #                               "run model.predict(X) on a couple examples"
        #                               "first")
        # else:
        scorers = ScorerFactory(self)
        return scorers

    @abc.abstractmethod
    def _execute(self, *args, **kwargs):
        """
        The way in which the submodule predicts values given an input
        """
        return


    @abc.abstractmethod
    def _predict(self, *args, **kwargs):
        """
        The way in which the submodule predicts values given an input
        """
        return

    @abc.abstractmethod
    def _get_static_predictor(self, *args, **kwargs):
        """Return a static prediction function to avoid shared state in multiprocessing"""
        return


    def __call__(self, *args, **kwargs):
        return self.predict(*args, **kwargs)


    def check_examples(self, examples):
        """
        Ties examples to self. equivalent to self.examples = np.array(examples).
        Parameters
        ----------
        examples: array type

        """
        if isinstance(examples, (pd.DataFrame, np.ndarray)):
            return examples
        else:
            return np.array(examples)


    def _if_no_prob(self, value):
        if self.probability == StaticTypes.unknown:
            return value
        else:
            return self.probability


    def _if_no_model(self, value):
        if self.model_type == StaticTypes.unknown:
            return value
        else:
            return self.model_type


    def _build_model_metadata(self, dataset):
        """
        Determines the model_type, output_type. Side effects
        of this method are to mutate object's attributes (model_type,
        n_classes, etc).

        Parameters
        ----------
        examples: pandas.DataFrame or numpy.ndarray
            The examples that will be passed through the predict function.
            The outputs from these examples will be used to make inferences
            about the types of outputs the function generally makes.

        """
        self.logger.debug("Beginning output checks")

        if self.input_type in (pd.DataFrame, None):
            outputs = self.predict(dataset.X)
        elif self.input_type == np.ndarray:
            outputs = self.predict(dataset.X)
        else:
            raise ValueError("Unrecognized input type: {}".format(self.input_type))

        self.input_shape = dataset.X.shape
        self.output_shape = outputs.shape

        ndim = len(outputs.shape)
        if ndim > 2:
            raise(ValueError("Unsupported model type, output dim = {}".format(ndim)))

        try:
            # continuous, binary, continuous multioutput, multiclass, multilabel-indicator
            self.output_type = type_of_target(outputs)
        except:
            self.output_type = False

        if self.output_type == 'continuous':
            # 1D array of continuous values
            self.model_type = self._if_no_model(StaticTypes.model_types.regressor)
            self.n_classes = 1
            self.probability = self._if_no_prob(False)

        elif self.output_type == 'multiclass':
            # 2D array of 1s and 0, exclusive
            self.model_type = self._if_no_model(StaticTypes.model_types.classifier)
            self.probability = self._if_no_prob(False)
            self.n_classes = len(np.unique(outputs))

        elif self.output_type == 'continuous-multioutput':
            # 2D array of continuous values
            self.model_type = self._if_no_model(StaticTypes.model_types.classifier)
            self.probability = self._if_no_prob(True)
            self.n_classes = outputs.shape[1]

        elif self.output_type == 'binary':
            # 2D array of 1s and 0, non exclusive
            self.model_type = self._if_no_model(StaticTypes.model_types.classifier)
            self.probability = self._if_no_prob(False)
            self.n_classes = 2

        elif self.output_type == 'multilabel-indicator':
            # 2D array of 1s and 0, non exclusive
            self.model_type = self._if_no_model(StaticTypes.model_types.classifier)
            self.probability = self._if_no_prob(False)
            self.n_classes = outputs.shape[1]

            if self.probability:
                self.output_type = 'continuous-multioutput'

        else:
            err_msg = "Could not infer model type"
            self.logger.debug("Inputs: {}".format(dataset.X))
            self.logger.debug("Outputs: {}".format(outputs))
            self.logger.debug("sklearn response: {}".format(self.output_type))
            exceptions.ModelError(err_msg)

        if self.target_names is None:
            self.target_names = ["predicted_{}".format(i) for i in range(self.n_classes)]

        if self.unique_values is None and self.model_type == 'classifier' and self.probability is False:
            raise (exceptions.ModelError('If using classifier without probability scores, unique_values cannot '
                                         'be None'))

        self.transformer = self.transformer_func_factory(outputs)

        reports = self.model_report(dataset.X)
        for report in reports:
            self.logger.debug(report)

        self.has_metadata = True


    def transformer_func_factory(self, outputs):
        """
        In the event that the predict func returns 1D array of predictions,
        then this returns a formatter to convert outputs to a 2D one hot encoded
        array.

        For instance, if:
            predict_fn(data) -> ['apple','banana']
        then
            transformer = Model.transformer_func_factory()
            transformer(predict_fn(data)) -> [[1, 0], [0, 1]]

        Returns
        ----------
        (callable):
            formatter function to wrap around predict_fn
        """

        # Note this expression below assumptions (not probability) evaluates to false if
        # and only if the model does not return probabilities. If unknown, should be true
        if self.model_type == StaticTypes.model_types.classifier and not self.probability:
            # fit label encoder
            # unique_values could ints/strings, etc.
            artificial_samples = np.array(self.unique_values)
            self.logger.debug("Label encoder fit on examples of shape: {}".format(outputs.shape))

            def check_classes(classes):
                if len(classes) > 2:
                    return classes
                elif len(classes) == 2:
                    # to get 2 columns from label_binarize, we need
                    # to pretend we have at least 3 classes.
                    # adding will preserve type of underlying classes
                    fake_class = classes[0] + classes[1]
                    return np.concatenate((classes, np.array([fake_class])))
                else:
                    raise ValueError("Less than 2 classes found in unique_classes")

            # defining this as a closure so it can be executed
            # outside the class
            def transformer(output):
                # numeric index of original classes.
                idx = list(range(len(artificial_samples)))
                classes = check_classes(artificial_samples)
                return label_binarize(output, classes)[:, idx]
            return transformer
        else:
            return identity_function


    def model_report(self, examples):
        """
        Just returns a list of model attributes as a list

        Parameters
        ----------
        examples: array type:
            Examples to use for which we report behavior of predict_fn.


        Returns
        ----------
        reports: list of strings
            metadata about function.

        """
        examples = DataManager(examples, feature_names=self.feature_names)
        reports = []
        if isinstance(self.examples, np.ndarray):
            raw_predictions = self.predict(examples)
            reports.append("Example: {} \n".format(examples[0]))
            reports.append("Outputs: {} \n".format(raw_predictions[0]))
        reports.append("Model type: {} \n".format(self.model_type))
        reports.append("Output Var Type: {} \n".format(self.output_var_type))
        reports.append("Output Shape: {} \n".format(self.output_shape))
        reports.append("N Classes: {} \n".format(self.n_classes))
        reports.append("Input Shape: {} \n".format(self.input_shape))
        reports.append("Probability: {} \n".format(self.probability))
        return reports

    def predict_subset_classes(self, data, subset_of_classes):
        """Filters predictions to a subset of classes."""
        if subset_of_classes is None:
            return self.predict(data)
        else:
            return DataManager(self.predict(data), feature_names=self.target_names)[subset_of_classes].X


def identity_function(x):
    return x
